/*
              *****************************************************************************
              *  Copyright (C) by Reinhard Heuberger                                 *
              *                                                                           *
              *  All rights reserved.                                                     *
              *                                                                           *
              *  Redistribution and use in source and binary forms, with or without       *
              *  modification, are permitted provided that the following conditions       *
              *  are met:                                                                 *
              *                                                                           *
              *  1. Redistributions of source code must retain the above copyright        *
              *     notice, this list of conditions and the following disclaimer.         *
              *  2. Redistributions in binary form must reproduce the above copyright     *
              *     notice, this list of conditions and the following disclaimer in the   *
              *     documentation and/or other materials provided with the distribution.  *
              *  3. Neither the name of the author nor the names of its contributors may  *
              *     be used to endorse or promote products derived from this software     *
              *     without specific prior written permission.                            *
              *                                                                           *
              *****************************************************************************
*/
//
// Description:
// ************
// FM ENcoder for RK05 disk emulator with 8 bit parallel input
// and serial output. In this case, the serial transfer clock runs
// at 1.44 Mhz. The clock input is 23.04 MHz, obtained from a PLL.
//
// TIMING:
// =======
// 5.76 MHz:  ....|-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_
// 2.88 MHz:  ....|--__--__--__--__--__--__--__--__--__--__--__--__
// 1.44 MHz:  ....|----____----____----____----____----____----____
// Pattern: 10101 |-___-___-_______-___-___-_______-___-___-_______
//                |    1   |   0   |   1   |   0   |   1   |
//
//                   Jan. 2023 
//
module ENcoder(
        CLK_FPGA,             // Input Clock  23.04 MHz
        Read_gate,
        data_byte_in,         // Data, 8 Bit parallel in
        //
        CLK_1_44,             // Out-Clock 1.44 MHZ
        CLK_2_88,
        CLK_5_76,
        FM,                   // FM out
        serial_OUT,           // Data, Serial out
        CLK_8bit,             // Out-clock, 8Bit 
        read_CLK);
//      
input  CLK_FPGA, Read_gate; 
input  [7:0] data_byte_in;
output CLK_1_44, CLK_2_88, CLK_5_76, FM;
output serial_OUT, CLK_8bit, read_CLK;
//
reg [31:0]  divider;         // counter,= clock generator.
reg [7:0]   shifter;         // 8Bit parallel in SR
reg [2:0] rclksync_16_4;     // like 2 D-FF register, 
reg CLK_1_44_p, CLK_8bit_P;  // Synchronized Clocks
reg FM, temp, nix; 
reg zero, one;
wire load;
//
initial
begin
   divider <= 0;
   shifter <= 0;
end
//
//
//
//=============== 8Bit load: Generate a 0.0434uS load pulse ==================
//============================================================================
always @ (posedge CLK_FPGA)
begin
   CLK_8bit_P <= CLK_8bit;
end
assign load = CLK_8bit & !CLK_8bit_P;
//
//========================= Mofify 1.44MHz clock  =============================
//============================================================================
always @ (posedge CLK_FPGA)
begin
    CLK_1_44_p <= CLK_1_44;
end
assign read_CLK = CLK_1_44 & !CLK_1_44_p;
//
//=============== Shiftregister with synchronous parallel load ===============
//========================== Counter/Divider =================================
//============================================================================
// Note: Counting up is done at negative edges.
//       Counting down is done at positive edges.
// RK05: FPGA Clock is 46.08MHz
//
always @ (posedge CLK_FPGA)
begin
    if(Read_gate) begin
        divider <= divider - 1;
        zero <=  ( divider[4] && divider[3] && divider[2] );
        one  <=  ( divider[4] && divider[3] && divider[2] ) | 
             (!divider[4] && divider[3] && divider[2] );
                //
        if( !CLK_1_44_p & CLK_1_44 ) begin           
            if (load ) begin
                shifter <= data_byte_in;             // LOAD SR
            end else if (!load) begin
                shifter <= shifter << 1;             // MSB first
                temp <= shifter[7]; 
            end
        end     
        //
        if (!temp) begin
            FM <= zero;                          // encode "0"
        end else if (temp) begin
            FM <= one;                           // encode "1"
        end
    end else begin
        nix <= nix;
    end
    // 
end
assign CLK_1_44    =  divider[4];  // Output: 1.44 MHz
assign CLK_2_88    =  divider[3];  // Output: 2.88 MHz
assign CLK_5_76    =  divider[2];  // Output: 5.76 MHz
assign CLK_8bit    =  divider[7];  // Output: 8Bit Clock
assign serial_OUT  =  temp; 
//
//
endmodule
